package fr.mastah;

import java.io.File;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.IOFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.maven.plugin.logging.Log;
import org.codehaus.plexus.util.StringUtils;

import fr.mastah.exception.BaseDirNotExistingException;
import fr.mastah.mojo.bean.JSDocInput;

/**
 * Class used to find the list of js files.
 * 
 * @author Mastah
 */
public class JSFileFinder {

	/** Logger. */
	private Log logger;

	/** File pattern matcher. */
	private Pattern pattern;

	/**
	 * Constructor.
	 * 
	 * @param theLogger logger.
	 */
	public JSFileFinder(final Log theLogger) {
		this.logger = theLogger;
	}

	/**
	 * Look for files in the base directories matching the pattern.
	 * 
	 * @param theJSDocInputs inputs containing a base directory and an optional pattern.
	 * @param theJSDocExclusions exclusions to remove from JSDoc3 process.
	 * @return list of files concatenated and separated by a space character.
	 * @throws BaseDirNotExistingException exception when one of the input base directory can't be found.
	 */
	public final String findJSFiles(final List<JSDocInput> theJSDocInputs, final List<String> theJSDocExclusions) throws BaseDirNotExistingException {

		// List of files that will be eligible for JSDoc documentation process
		Collection<File> eligibleFiles = new HashSet<File>();

		// Directory in which we will look for files
		File jsDir = null;

		this.logger.info("");

		// For each JSDocInput
		for (JSDocInput currentInput : theJSDocInputs) {

			// We check for the value of 'jsDir'
			if (StringUtils.isNotEmpty(currentInput.getJsDir())) {

				// We create a file from the jsDir value
				jsDir = new File(currentInput.getJsDir().replace("\\", "/"));

				// Creating the pattern that will be used to match for files
				this.pattern = Pattern.compile(currentInput.getPattern());

				// If the base dir doesn't exist and is not a directory then we throw an exception
				if (!jsDir.exists() || !jsDir.isDirectory()) {
					throw new BaseDirNotExistingException("Could not find the base directory of files to scan: " + currentInput.getJsDir()
					                                      + " does not exists or is not a directory.");
				}

				this.logger.info("Looking for files in " + jsDir.getPath() + " with pattern " + currentInput.getPattern());

				// File filter
				IOFileFilter fileFilter = new IOFileFilter() {
					@Override
					public boolean accept(final File theDir, final String theFileName) {
						return false;
					}

					@Override
					public boolean accept(final File theDir) {

						// Result
						Boolean result = Boolean.FALSE;

						final Boolean isFile = theDir.isFile();
						final Boolean isMatchFile = JSFileFinder.this.pattern.matcher(theDir.getPath()).find();
						final Boolean isExclude = theJSDocExclusions.contains(theDir.getName());

						// If it's a file, the file match the pattern and the file is not excluded then the file is eligible.
						if (isFile && isMatchFile && !isExclude) {
							result = Boolean.TRUE;
						}
						return result;
					}
				};

				// We add all eligible files to the result files
				eligibleFiles.addAll(FileUtils.listFiles(jsDir, fileFilter, TrueFileFilter.INSTANCE));

			} else {
				throw new BaseDirNotExistingException("'jsDir' can't be empty.");
			}

		}

		// We convert the eligible files to a String representing all the files concatenated with a space
		return this.buildListOfFileAsString(eligibleFiles);
	}

	/**
	 * Transform a list of File to a String concatenating the files paths.
	 * 
	 * @param theListOfFiles the list of files.
	 * @return string representing the list of files.
	 */
	private String buildListOfFileAsString(final Collection<File> theListOfFiles) {

		this.logger.info("");
		this.logger.info("Files to process:");

		// Result
		StringBuilder build = new StringBuilder();

		Iterator<File> xIt = theListOfFiles.iterator();
		File currentFile = null;
		while (xIt.hasNext()) {
			currentFile = xIt.next();
			this.logger.info("	+ " + currentFile.getPath());
			build.append(currentFile.getPath());
			if (xIt.hasNext()) {
				build.append(" ");
			}
		}

		return build.toString();
	}

}
