package fr.mastah.mojo;

import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;

import fr.mastah.ExtractJSDoc3;
import fr.mastah.JSFileFinder;
import fr.mastah.exception.BaseDirNotExistingException;
import fr.mastah.exception.JreDirNotExistingException;
import fr.mastah.exception.TmpDirNotExistingException;
import fr.mastah.mojo.bean.JSDocInput;
import fr.mastah.mojo.process.ErrorInputStreamHandler;
import fr.mastah.mojo.process.InfoInputStreamHandler;

/**
 * @author Mastah
 */
@Mojo(name = "jsdoc3", defaultPhase = LifecyclePhase.COMPILE)
public class JSDoc3Mojo extends AbstractMojo {

	/**
	 * Command line exec for JSDoc3.
	 * 
	 * <pre>
	 *    {0} : java exec dir
	 *    {1} : jsdoc tmp dir
	 *    {2} : result directory
	 *    {3} : list of files
	 *    {4} : jsdoc3 options
	 * </pre>
	 */
	private static final String RAW_COMMAND_LINE = "{0} -classpath {1}/rhino/js.jar org.mozilla.javascript.tools.shell.Main "
	                                               + "-modules file:///{1}/modules -modules file:///{1}/rhino "
	                                               + "-modules file:///{1}/lib {1}/jsdoc/jsdoc.js -d {2} {3} {4} --dirname={1}";

	/** Logger. */
	private Log logger;

	/** Output directory to be used for JSDoc3. */
	@Parameter(property = "jsdoc3.outputDirectory", defaultValue = "${project.build.directory}\\jsdoc3")
	private String outputDirectory;

	/**
	 * Inputs that will be used by JSDoc3 for documentations.
	 * 
	 * <pre>
	 * Format as follow:
	 * 	&lt;jsdocInputs&gt;
	 * 		&lt;jsdocInput&gt;
	 * 			...
	 * 		&lt;/jsdocInput&gt;
	 * 	&lt;/jsdocInputs&gt;
	 * </pre>
	 */
	@Parameter(property = "jsdoc3.jsdocInputs", required = true)
	private List<JSDocInput> jsdocInputs;

	/** Files to remove from the JSDoc3 processing. */
	@Parameter(property = "jsdoc3.jsdocExclusions")
	private List<String> jsdocExclusions = new ArrayList<String>();

	/** Direct JSDoc3 options. */
	@Parameter(property = "jsdoc3.options", defaultValue = "")
	private String options = "";

	/** Used to extract JSDoc3 files. */
	private ExtractJSDoc3 extractJSDoc3;

	/**
	 * Constructor.
	 */
	public JSDoc3Mojo() {
		super();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void execute() throws MojoExecutionException, MojoFailureException {

		// Retrieve lastest logger
		this.logger = this.getLog();

		// Init classes
		this.extractJSDoc3 = new ExtractJSDoc3();

		// Create the class util that will look for the js files
		final JSFileFinder jsFileFinder = new JSFileFinder(this.logger);

		try {

			this.generateLogo();

			this.logger.info("JSDoc3 documentations will be created into: " + this.outputDirectory);

			// Look for js files matching the inputs directories and patterns.
			String jsFiles = jsFileFinder.findJSFiles(this.jsdocInputs, this.jsdocExclusions);

			// Fetch tmpDir
			String tmpDir = this.locateTmpDir();

			// Extract JSDoc3
			this.logger.info("");
			this.logger.info("Extracting JSDoc3 to: " + tmpDir);
			this.extractJSDoc3.extract(tmpDir);

			// Build command line for JSDoc3
			String commandLine = this.buildExecCommandLine(tmpDir, jsFiles);

			// Execute JSDoc3
			this.logger.info("");
			this.logger.info("Executing JSDoc3:\n" + commandLine);
			Process p = Runtime.getRuntime().exec(commandLine);

			InputStream inStream = p.getInputStream();
			Thread tInfo = new InfoInputStreamHandler(inStream, this.logger);

			InputStream errStream = p.getErrorStream();
			Thread tError = new ErrorInputStreamHandler(errStream, this.logger);

			p.waitFor();

			while (tInfo.isAlive() || tError.isAlive()) {
				// Do nothing, we wait for the process of JSDoc3 to end
			}

			if (p.exitValue() != 0) {
				throw new MojoFailureException("JSDoc3 failed.");
			}

		} catch (IOException e) {
			throw new MojoExecutionException("Could not extract JSDoc3 files.", e);
		} catch (InterruptedException e) {
			throw new MojoExecutionException("Could not execute JSDoc3.", e);
		} catch (TmpDirNotExistingException e) {
			throw new MojoExecutionException("Could not find tmp dir on your system.", e);
		} catch (JreDirNotExistingException e) {
			throw new MojoExecutionException("Could not find jre dir on your system.", e);
		} catch (BaseDirNotExistingException e) {
			throw new MojoExecutionException("Error while looking for file to scan for documentation.", e);
		}
	}

	/**
	 * Locate the tmp dir on the system.
	 * 
	 * @return tmp dir path of the system.
	 * @throws TmpDirNotExistingException exception when no tmp dir could be found.
	 */
	private String locateTmpDir() throws TmpDirNotExistingException {
		// Fetch tmp dir
		String result = System.getProperty("java.io.tmpdir");

		// If no tmp dir could be found
		if ((result == null) || "".equals(result)) {
			throw new TmpDirNotExistingException();
		}

		return result;
	}

	/**
	 * Build the command line.
	 * 
	 * @param theTmpDir the tmp dir where jsdoc will be extraced.
	 * @param theJSFilesToDocument the list of js files to document via JSDoc.
	 * @return the command line corresponding to the generation of documentation.
	 * @throws JreDirNotExistingException exception when no jre dir could be found.
	 */
	private String buildExecCommandLine(final String theTmpDir, final String theJSFilesToDocument) throws JreDirNotExistingException {

		// Result cmd
		String result;

		// Fetch jre home
		String jerHome = System.getProperty("java.home");

		// If no jre dir could be found
		if ((jerHome == null) || "".equals(jerHome)) {
			throw new JreDirNotExistingException();
		}

		// Define java exec
		String javaExec = jerHome + "/bin/java";

		// Define JSDoc3 tmp folder
		String jsdoc3Dir = theTmpDir + "/jsdoc3";

		// Format the command line
		result = MessageFormat.format(JSDoc3Mojo.RAW_COMMAND_LINE, javaExec, jsdoc3Dir, this.outputDirectory, theJSFilesToDocument, this.options);
		result = result.replace("\\", "/");

		return result;
	}

	/**
	 * Setter for outputDirectory.
	 * 
	 * @param theOutputDirectory the outputDirectory to set.
	 */
	public final void setOutputDirectory(final String theOutputDirectory) {
		this.outputDirectory = theOutputDirectory;
	}

	/**
	 * Setter for jsdocInputs.
	 * 
	 * @param theJsdocInputs the jsdocInputs to set.
	 */
	public final void setJsdocInputs(final List<JSDocInput> theJsdocInputs) {
		this.jsdocInputs = theJsdocInputs;
	}

	/**
	 * Setter for jsdocExclusions.
	 * 
	 * @param theJsdocExclusions the jsdocExclusions to set.
	 */
	public final void setJsdocExclusions(final List<String> theJsdocExclusions) {
		this.jsdocExclusions = theJsdocExclusions;
	}

	/**
	 * Setter for options.
	 * 
	 * @param theOptions the options to set.
	 */
	public final void setOptions(final String theOptions) {
		this.options = theOptions;
	}

	/**
	 * Generate JSDoc3 logo.
	 */
	public final void generateLogo() {
		this.logger.info("");
		this.logger.info("		       _______ ____            _____");
		this.logger.info("		      / / ___// __ \\____  ____|__  /");
		this.logger.info("		 __  / /\\__ \\/ / / / __ \\/ ___//_ < ");
		this.logger.info("		/ /_/ /___/ / /_/ / /_/ / /_____/ / ");
		this.logger.info("		\\____//____/_____/\\____/\\___/____/  ");
		this.logger.info("");
	}

}
